@@ -61,9 +61,15 @@
     Updated by      Richard Gooch   18-JUN-1999: Added "OPTIONAL_INCLUDE" key.
   Added ignore actions. Added event mask.
 
-    Last updated by Richard Gooch   24-JUN-1999: Switched to separated event
+    Updated by      Richard Gooch   24-JUN-1999: Switched to separated event
   and action config file specifiers.
 
+    Updated by      Richard Gooch   20-OCT-1999: Fixed typo in debugging output
+
+    Last updated by Richard Gooch   24-OCT-1999: Added "-fg" switch and
+  redefined tracing levels. Delay opening syslog until /dev/log appears. Scan
+  mounted FS and generate synthetic REGISTER events.
+
 
 */
 #include <stdio.h>
@@ -88,7 +94,7 @@
 #include <karma.h>
 
 
-#define VERSION "1.2.3"
+#define VERSION "1.2.4"
 #define CONFIG_FILE "/etc/devfsd.conf"
 #define MAX_ARGS (6 + 1)
 
@@ -108,9 +114,8 @@
 #define MKACTION(what,when) (struct action_type) {what, when}
 
 #define SYSLOG(level, format, args...) \
-    {if ( debug_protocol || (trace_level > 0) ) \
-         fprintf (stderr, format, ## args); \
-     else syslog (level, format, ##args); \
+    {if (syslog_is_open) syslog (level, format, ##args); \
+     else fprintf (stderr, format, ## args); \
     }
 
 
@@ -178,12 +183,16 @@ static gid_t get_gid (const char *string);
 static mode_t get_mode (const char *string);
 static void sig_handler (int sig);
 static CONST char *get_variable (CONST char *variable, void *info);
+static void do_open_syslog ();
+static void do_scan_and_service (DIR *dp, CONST char *dirname);
 
 
 /*  Private data  */
 static struct config_entry *first_config = NULL;
 static struct config_entry *last_config = NULL;
 static const char *mount_point = NULL;
+static flag no_syslog = FALSE;
+static flag syslog_is_open = FALSE;
 static char trace_level = 0;
 static flag debug_protocol = FALSE;
 static struct initial_symlink_struct
@@ -206,12 +215,14 @@ int main (int argc, char **argv)
 {
     flag print_version = FALSE;
     flag debug_devfs = FALSE;
+    flag do_daemon = TRUE;
     int fd, proto_rev, count;
     unsigned int devfs_debug_mask;
+    unsigned long event_mask = 0;
     struct sigaction new_action;
     struct stat statbuf;
     char filename[STRING_LENGTH];
-    static char usage[] = "devfsd mntpnt [-v] [-d] [-t num] [-D mask]";
+    static char usage[] = "devfsd mntpnt [-v] [-d] [-t num] [-D mask] [-fg]";
 
     if (argc < 2)
     {
@@ -221,22 +232,26 @@ int main (int argc, char **argv)
     for (count = 2; count < argc; ++count)
     {
 	if (strcmp (argv[count], "-v") == 0) print_version = TRUE;
-	else if (strcmp (argv[count], "-d") == 0) debug_protocol = TRUE;
-	else if ( (strcmp (argv[count], "-t") == 0) && (count + 1 < argc) )
+	else if (strcmp (argv[count], "-d") == 0)
 	{
-	    trace_level = atoi (argv[++count]);
+	    debug_protocol = TRUE;
+	    no_syslog = TRUE;
 	}
+	else if ( (strcmp (argv[count], "-t") == 0) && (count + 1 < argc) )
+	    trace_level = atoi (argv[++count]);
 	else if ( (strcmp (argv[count], "-D") == 0) && (++count < argc) )
 	{
 	    devfs_debug_mask = strtol (argv[count], NULL, 0);
 	    debug_devfs = TRUE;
 	}
+	else if (strcmp (argv[count], "-fg") == 0) do_daemon = FALSE;
 	else
 	{
 	    fprintf (stderr, "Usage:\t%s\n", usage);
 	    exit (1);
 	}
     }
+    if (trace_level > 0) no_syslog = TRUE;
     mount_point = argv[1];
     strcpy (filename, argv[1]);
     strcat (filename, "/.devfsd");
@@ -268,7 +283,7 @@ int main (int argc, char **argv)
 	    exit (0);
 	}
     }
-    if ( print_version || (trace_level > 0) || debug_protocol ||
+    if ( print_version || (trace_level > 1) || debug_protocol ||
 	 (DEVFSD_PROTOCOL_REVISION_DAEMON != proto_rev) )
     {
 	fprintf (stderr, "devfsd: Linux device filesystem daemon v%s\n",
@@ -299,15 +314,24 @@ int main (int argc, char **argv)
     new_action.sa_flags = 0;
     /*  Set up SIGHUP handler  */
     new_action.sa_handler = sig_handler;
-    if (sigaction (SIGHUP, &new_action, (struct sigaction *) NULL) != 0)
+    if (sigaction (SIGHUP, &new_action, NULL) != 0)
     {
 	fprintf (stderr, "Error setting SIGHUP handler\t%s\n", ERRSTRING);
 	exit (1);
     }
     if (trace_level < 1)
-    {
     	fprintf (stderr, "Started device management daemon for %s\n",
 		 mount_point);
+    else
+    	fprintf (stderr,
+		 "Started device management daemon for %s at trace level %d\n",
+		 mount_point, trace_level);
+    read_config (CONFIG_FILE, FALSE, &event_mask);
+    /*  Do the scan before forking, so that boot scripts see the finished
+	product  */
+    do_scan_and_service (opendir (mount_point), mount_point);
+    if (do_daemon)
+    {
 	switch ( fork () )
 	{
 	  case 0:
@@ -323,25 +347,13 @@ int main (int argc, char **argv)
 	    exit (0);
 	    /*break;*/
 	}
-	openlog ("devfsd", LOG_PID, LOG_DAEMON);
-	close (0);
-	close (1);
-	close (2);
-    }
-    else
-    {
-    	fprintf (stderr,
-		 "Started device management daemon for %s at trace level %d\n",
-		 mount_point, trace_level);
     }
+    do_open_syslog ();
     while (TRUE)
     {
-	unsigned long event_mask = 0;
-
-	read_config (CONFIG_FILE, FALSE, &event_mask);
-	/*  Carry on daemon  */
 	do_servicing (fd, event_mask);
 	free_config ();
+	read_config (CONFIG_FILE, FALSE, &event_mask);
     }
 }   /*  End Function main  */
 
@@ -382,7 +394,7 @@ static void read_config (CONST char *filename, flag optional,
 		CONFIG_FILE, ERRSTRING);
 	exit (1);
     }
-    if (trace_level > 1) fprintf (stderr, "Reading file: \"%s\"\n",
+    if (trace_level > 2) fprintf (stderr, "Reading file: \"%s\"\n",
 				  CONFIG_FILE);
     while (fgets (buf, STRING_LENGTH, fp) != NULL)
     {
@@ -510,7 +522,8 @@ static void read_config (CONST char *filename, flag optional,
 	last_config = new;
     }
     fclose (fp);
-    SYSLOG (LOG_INFO, "read config file: \"%s\"\n", CONFIG_FILE);
+    if (syslog_is_open)
+	SYSLOG (LOG_INFO, "read config file: \"%s\"\n", CONFIG_FILE);
 }   /*  End Function read_config   */
 
 static void do_servicing (int fd, unsigned long event_mask)
@@ -522,9 +535,16 @@ static void do_servicing (int fd, unsigned long event_mask)
 {
     ssize_t bytes;
     struct devfsd_notify_struct info;
+    unsigned long tmp_event_mask;
 
     /*  Tell devfs what events we care about  */
-    if (ioctl (fd, DEVFSDIOC_SET_EVENT_MASK, event_mask) != 0)
+    tmp_event_mask = event_mask;
+    /*  May need to trap inode creates to watch for syslogd(8) starting  */
+    if (!syslog_is_open && !no_syslog)
+    {
+	tmp_event_mask |= 1 << DEVFSD_NOTIFY_CREATE;
+    }
+    if (ioctl (fd, DEVFSDIOC_SET_EVENT_MASK, tmp_event_mask) != 0)
     {
 	SYSLOG (LOG_ERR, "error setting event mask\t%s\n", ERRSTRING);
 	exit (1);
@@ -536,6 +556,19 @@ static void do_servicing (int fd, unsigned long event_mask)
 	    SYSLOG (LOG_ERR, "devfs closed file!\n");
 	    exit (1);
 	}
+	/*  Special trap for "/dev/log" creation  */
+	if (!syslog_is_open && !no_syslog &&
+	    (info.type == DEVFSD_NOTIFY_CREATE) &&
+	    (strcmp (info.devname, "log") == 0) )
+	{
+	    /*  Open syslog, now that "/dev/log" exists  */
+	    do_open_syslog ();
+	    if (ioctl (fd, DEVFSDIOC_SET_EVENT_MASK, event_mask) != 0)
+	    {
+		SYSLOG (LOG_ERR, "error setting event mask\t%s\n", ERRSTRING);
+		exit (1);
+	    }
+	}
 	service_name (&info);
     }
     if (errno == EINTR) return;
@@ -554,11 +587,11 @@ static void service_name (const struct devfsd_notify_struct *info)
 
     if (info->overrun_count > 0)
 	SYSLOG (LOG_ERR, "%u events have been lost!\n", info->overrun_count);
-    if (trace_level > 1) fprintf (stderr, "Looking for \"%s\" (%d)\n",
+    if (trace_level > 2) fprintf (stderr, "Looking for \"%s\" (%d)\n",
 				  info->devname, info->type);
     for (entry = first_config; entry != NULL; entry = entry->next)
     {
-	if (trace_level > 2) fprintf (stderr, "\t\tProcess \"%s\" (%d)\n",
+	if (trace_level > 3) fprintf (stderr, "\t\tProcess \"%s\" (%d)\n",
 				      info->devname, entry->action.when);
 	/*  First check if action matches the type, then check if name matches
 	 */
@@ -600,16 +633,15 @@ static void update_permissions (const struct devfsd_notify_struct *info,
     strcat (path, info->devname);
     if (lstat (path, &statbuf) != 0)
     {
-	SYSLOG (LOG_ERR, "error stat(2)ing: \"%s\"\t%s\n",
-		path, ERRSTRING);
+	SYSLOG (LOG_ERR, "error stat(2)ing: \"%s\"\t%s\n", path, ERRSTRING);
 	SYSLOG (LOG_ERR, "exiting\n");
 	exit (1);
     }
     new_mode = (statbuf.st_mode & S_IFMT) |
 	(entry->u.permissions.mode & ~S_IFMT);
-    if (trace_level > 1)
+    if (trace_level > 2)
 	fprintf (stderr,
-		 "\tupdate permissions for \"%s\" from %05o to %05o, user.goup from %d.%d to %d.%d\n",
+		 "\tupdate permissions for \"%s\" from %05o to %05o, user.group from %d.%d to %d.%d\n",
 		 path, info->mode, new_mode, info->uid, info->gid,
 		 entry->u.permissions.uid, entry->u.permissions.gid);
     if (new_mode != statbuf.st_mode)
@@ -621,7 +653,7 @@ static void update_permissions (const struct devfsd_notify_struct *info,
 	    SYSLOG (LOG_ERR, "exiting\n");
 	    exit (1);
 	}
-	if (trace_level > 1) fprintf (stderr,"Set permission %05o on \"%s\"\n",
+	if (trace_level > 2) fprintf (stderr,"Set permission %05o on \"%s\"\n",
 				      new_mode, path);
     }
     if ( (entry->u.permissions.uid != statbuf.st_uid) ||
@@ -635,7 +667,7 @@ static void update_permissions (const struct devfsd_notify_struct *info,
 	    SYSLOG (LOG_ERR, "exiting\n");
 	    exit (1);
 	}
-	if (trace_level > 1)
+	if (trace_level > 2)
 	    fprintf (stderr, "Set user.group %d.%d on \"%s\"\n",
 		     entry->u.permissions.uid, entry->u.permissions.gid, path);
     }
@@ -659,7 +691,7 @@ static void execute (const struct devfsd_notify_struct *info,
     strcpy (gv_info.devpath, mount_point);
     strcat (gv_info.devpath, "/");
     strcat (gv_info.devpath, info->devname);
-    if (trace_level > 0) fprintf (stderr, "Calling %s for \"%s\" | \"%s\":",
+    if (trace_level > 1) fprintf (stderr, "Calling %s for \"%s\" | \"%s\":",
 				  entry->u.execute.argv[0], info->devname,
 				  gv_info.devpath);
     for (count = 0; entry->u.execute.argv[count] != NULL; ++count)
@@ -668,10 +700,10 @@ static void execute (const struct devfsd_notify_struct *info,
 			entry->u.execute.argv[count], get_variable, &gv_info,
 			NULL);
 	argv[count] = largv[count];
-	if (trace_level > 0) fprintf (stderr, " '%s'", argv[count]);
+	if (trace_level > 1) fprintf (stderr, " '%s'", argv[count]);
     }
-    if (trace_level > 0) fprintf (stderr, ". i.e. %d arg%s\n",
-				  count-1, (count == 2) ? "" : "s");
+    if (trace_level > 1) fprintf (stderr, ". i.e. %d arg%s\n",
+				  count - 1, (count == 2) ? "" : "s");
     argv[count] = NULL;
     switch ( fork () )
     {
@@ -843,7 +875,7 @@ static mode_t get_mode (const char *string)
     if (string[6] == 'r') mode |= S_IROTH;
     if (string[7] == 'w') mode |= S_IWOTH;
     if (string[8] == 'x') mode |= S_IXOTH;
-    if (trace_level > 2) fprintf (stderr, "\tMode \"%s\" gives %05o\n",
+    if (trace_level > 3) fprintf (stderr, "\tMode \"%s\" gives %05o\n",
 				  string, mode);
     return (mode);
 }   /*  End Function get_mode  */
@@ -891,3 +923,57 @@ static CONST char *get_variable (CONST char *variable, void *info)
     }
     return (NULL);
 }   /*  End Function get_variable  */
+
+static void do_open_syslog ()
+/*  [SUMMARY] Open syslog(3) connection if appropriate.
+    [RETURNS] Nothing.
+*/
+{
+    if (syslog_is_open || no_syslog) return;
+    if (access ("/dev/log", F_OK) != 0) return;
+    openlog ("devfsd", LOG_PID, LOG_DAEMON);
+    close (0);
+    close (1);
+    close (2);
+    syslog_is_open = TRUE;
+}   /*  End Function do_open_syslog  */
+
+static void do_scan_and_service (DIR *dp, CONST char *dirname)
+/*  [SUMMARY] Scan a directory tree and generate register events on leaf nodes.
+    <dp> The directory pointer.
+    <dirname> The name of the directory.
+    [RETURNS] Nothing.
+*/
+{
+    struct stat statbuf;
+    struct dirent *de;
+    char path[STRING_LENGTH];
+
+    while ( (de = readdir (dp) ) != NULL )
+    {
+	struct devfsd_notify_struct info;
+
+	if ( (strcmp (de->d_name, ".") == 0) ||
+	     (strcmp (de->d_name, "..") == 0) ) continue;
+	sprintf (path, "%s/%s", dirname, de->d_name);
+	if (lstat (path, &statbuf) != 0)
+	{
+	    SYSLOG (LOG_ERR, "Error stating file \"%s\"\t%s\n",
+		    path, ERRSTRING);
+	    exit (RV_SYS_ERROR);
+	}
+	if ( S_ISDIR (statbuf.st_mode) )
+	{
+	    do_scan_and_service (opendir (path), path);
+	    continue;
+	}
+	memset (&info, 0, sizeof info);
+	info.type = DEVFSD_NOTIFY_REGISTERED;
+	info.mode = statbuf.st_mode;
+	info.uid = statbuf.st_uid;
+	info.gid = statbuf.st_gid;
+	strcpy (info.devname, path + strlen (mount_point) + 1);
+	info.namelen = strlen (info.devname);
+	service_name (&info);
+    }
+}   /*  End Function do_scan_and_service  */
